import java.awt.*;
import java.awt.event.*;
import javax.swing.JComponent;

@SuppressWarnings("serial")
public class GraphPanel extends JComponent {
    private Dimension size = new Dimension();
    private int x[], xmax, xmin;
    private int y[], ymax, ymin;
    private GraphApplication caller;

    /** XOR 描画を正確にするために、バッファリングする */
    private Dimension offDimension = new Dimension();
    private Image offImage;
    private Graphics offGraphics;
    private boolean needUpdate = false;

    /** 点の初期色 */
    final static Color DEFAULT_NODE_COLOR = Color.green;
    /** 枝の初期色 */
    final static Color DEFAULT_EDGE_COLOR = Color.orange;
    /** 点の大きさ */
    final static int RADIUS = 3;

    /** 点の色 */
    private Color nodeColor = DEFAULT_NODE_COLOR;
    /** 枝の色 */
    private Color edgeColor = DEFAULT_EDGE_COLOR;
    /** 枝の背景色(null でなければ XOR 描画する) */
    private Color edgeBgColor = null;

    /** マウスで頂点を選ぶ(選択中の頂点を目立つように描画する) */
    public boolean mouseHandle = false;
    /** マウスで頂点を選んでいる最中 */
    public boolean capturing = false;
    /** 選んでいる頂点のインデックス */
    public int selectedIndex = 0;
    //public boolean clicked = false;

    public GraphPanel( GraphApplication caller ) {
        this.caller = caller;
        Mouse m = new Mouse();
        addMouseListener(m);
        setBackground(GraphApplication.BG_COLOR);
    }

    public void setColors( Color node, Color edge, Color edgeBg ) {
        nodeColor = (node == null) ? DEFAULT_NODE_COLOR : node;
        edgeColor = (edge == null) ? DEFAULT_EDGE_COLOR : edge;
        edgeBgColor = edgeBg;
    }

    public void setNodes(int[] x, int[] y) {
        this.x = x;
        this.y = y;
        if (x.length != y.length) {
            throw new IllegalArgumentException("x.length != y.length");
        }
        xmax = ymax = Integer.MIN_VALUE;
        xmin = ymin = Integer.MAX_VALUE;
        for (int i = 0; i < x.length; i++) {
            if (xmax < x[i]) xmax = x[i];
            if (xmin > x[i]) xmin = x[i];
            if (ymax < y[i]) ymax = y[i];
            if (ymin > y[i]) ymin = y[i];
        }
    }

    private int screenX(int x) {
        return 20 + (size.width - 40) * (x - xmin) / (xmax - xmin);
    }

    private int screenY(int y) {
        return 20 + (size.height - 40) * (y - ymax) / (ymin - ymax);
    }

    public void drawEdge( int i, int j ) {
        Graphics g = offGraphics;
        if ( g == null ) return; // 起こらないはず
        synchronized (offGraphics) {
            g.setColor( edgeColor );
            if ( edgeBgColor == null ) g.setPaintMode();
            else                       g.setXORMode( edgeBgColor );
            drawEdge( i, j, g );
        }
    }

    public synchronized void drawEdge( int i, int j, Graphics g ) {
        if ( i > j ) {  int swap = i;  i = j;  j = swap;  }
        g.drawLine( screenX(x[i]), screenY(y[i]),
                screenX(x[j]), screenY(y[j]) );
        needUpdate = true;
    }

    private void drawNode( int index, Graphics g ) {
        g.fillArc( screenX(x[index])-RADIUS,
                screenY(y[index])-RADIUS,
                RADIUS*2+1, RADIUS*2+1, 0, 360 );
        needUpdate = true;
    }
    
    public void drawNodeCircle( int index ) {
        Graphics g = offGraphics;
        if ( g == null ) return; // 起こらないはず
        synchronized (offGraphics) {
            drawNodeCircle( index, g );
        }
    }

    private void drawNodeCircle( int index, Graphics g ) {
        g.setColor( edgeColor );
        g.setXORMode( getBackground() );
        g.fillArc( screenX(x[index])-RADIUS*2,
                   screenY(y[index])-RADIUS*2,
                   RADIUS*4+1, RADIUS*4+1, 0, 360 );
        needUpdate = true;
    }

    public void repaintIfNeeded() {
        if (needUpdate) repaint();
    }
    
    @Override
    public void paintComponent( Graphics g ) {
        getSize(size);
        if (makeOffGraphics(size)) updateOffGraphics();
        g.drawImage(offImage, 0, 0, this);
        needUpdate = false;
    }

    private boolean makeOffGraphics(Dimension size) {
        if ( (offGraphics == null)
                || (size.width != offDimension.width)
                || (size.height != offDimension.height) ) {
            offDimension.setSize(size);
            offImage = createImage(size.width, size.height);
            offGraphics = offImage.getGraphics();
            //System.out.println("OffGraphics created.");
            return true;
        }
        return false;
    }

    public void updateOffGraphics() {
        if (offGraphics == null) return;
        offGraphics.clearRect(0, 0, offDimension.width, offDimension.height);
        if ( x == null ) return;
        System.out.println("updateOffGraphics");
        
        synchronized (offGraphics) {
            offGraphics.setColor( edgeColor );
            if ( edgeBgColor == null ) offGraphics.setPaintMode();
            else                       offGraphics.setXORMode( edgeBgColor );
            caller.drawAllEdges( offGraphics );
            
            offGraphics.setColor( nodeColor );
            offGraphics.setPaintMode(); // XOR mode をやめる
            for ( int i=0; i<x.length; i++ ) drawNode( i, offGraphics );

            if ( mouseHandle ) drawNodeCircle( selectedIndex, offGraphics );
        }
    }
    
    private int getNearestIndex( int mouseX, int mouseY ) {
        int index = -1;
        int dst2 = Integer.MAX_VALUE;
        for (int i=0; i<x.length; i++) {
            int dx = mouseX - screenX(x[i]);
            int dy = mouseY - screenY(y[i]);
            int tmp = dx*dx + dy*dy;
            if (dst2 > tmp) {
                dst2 = tmp;
                index = i;
            }
        }
        return index;
    }

    public void setMouseHandle( boolean enable ) {
        if ( mouseHandle ^ enable ) drawNodeCircle( selectedIndex );
        mouseHandle = capturing = enable;
    }

    private class Mouse extends MouseAdapter {
        public void mouseClicked(MouseEvent e) {
            if (!capturing) return;
            int i = getNearestIndex(e.getX(), e.getY());
            if (selectedIndex != i) {
                drawNodeCircle( i );
                drawNodeCircle( selectedIndex );
                selectedIndex = i;
            }
        }
    }
}
